/*********************************************************************
 * Desktop Station Core Library for Arduino
 *
 * Copyright (C) 2014 Yaasan
 * Not use for commercial use
 */

#include <arduino.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DS_Type.h"
#include "DS_Common.h"
#include "DS_List.h"
#include "DS_DCC.h"
#include "DS_MM2.h"
#include "DSCore.h"


DS_ITEM gList[ITEM_LISTSIZE];/**< 送信管理用リスト */
UC gScanIndex = 0;
boolean gPoweron = false;

DSCoreLib::DSCoreLib()
{
	
	gPoweron = false;
	
}


UC DSCoreLib::MAIN_scanList(DS_ITEM *inList, UC inScanIndex)
{
	UC aRetIndex = inScanIndex;
	UC aMM2_Len;
	
	if( inScanIndex >= ITEM_LISTSIZE)
	{
		return 255;
	}
	
	/* パケットを送信する */
	if( inList[inScanIndex].mType == ITEMTYPE_MM2BASE )
	{
		/* 前回のパケットと区別するためにウェイト期間 */
		_delay_ms(6);

		/* 拡張ファンクションかどうかチェックする */
		if(inList[inScanIndex].mFuncNo > 5 )
		{
			/* 拡張ファンクション信号(F5-F16)を送る。Double Packet x3 (6回)単位。速度はゼロ。 */
			MM2_SendDoublePackets(inList[inScanIndex].mDatas, 3, MODEMM2_BASE, inList[inScanIndex].mFuncNo, inList[inScanIndex].mFunctionBuf);
			MM2_WaitLong();
			MM2_SendDoublePackets(inList[inScanIndex].mDatas, 3, MODEMM2_BASE, inList[inScanIndex].mFuncNo, inList[inScanIndex].mFunctionBuf);
			MM2_WaitLong();
			MM2_SendDoublePackets(inList[inScanIndex].mDatas, 3, MODEMM2_BASE, inList[inScanIndex].mFuncNo, inList[inScanIndex].mFunctionBuf);
			
		}
		else
		{
			/* 速度信号のみ送る(2x double packet) */
			MM2_SendDoublePackets(inList[inScanIndex].mDatas, 3, MODEMM2_BASE, 0, inList[inScanIndex].mFunctionBuf);
			MM2_WaitLong();
			MM2_SendDoublePackets(inList[inScanIndex].mDatas, 3, MODEMM2_BASE, 0, inList[inScanIndex].mFunctionBuf);

			if((inList[inScanIndex].mLen > 3) || (inList[inScanIndex].mFuncNo > 0) )
			{
				aMM2_Len = inList[inScanIndex].mLen > 3 ? 4 : 3;
			
				/* 50ms待つ */
				_delay_ms(50);
				
				/* ファンクション信号、28Step信号、ファンクション＆28Step信号を送る */
				MM2_SendDoublePackets(inList[inScanIndex].mDatas, aMM2_Len, MODEMM2_BASE, inList[inScanIndex].mFuncNo, inList[inScanIndex].mFunctionBuf);
				MM2_WaitLong();
				MM2_SendDoublePackets(inList[inScanIndex].mDatas, aMM2_Len, MODEMM2_BASE, inList[inScanIndex].mFuncNo, inList[inScanIndex].mFunctionBuf);
		
			}
		
		}
				
		/* カウンタをデクリメントする */
		if( inList[inScanIndex].mFuncCounter > 0)
		{
			inList[inScanIndex].mFuncCounter--;
		}
		
		/* ファンクション出力を停止 */
		if((inList[inScanIndex].mFuncCounter == 0) && (inList[inScanIndex].mFuncNo > 0))
		{
			inList[inScanIndex].mFuncNo = 0;
		}
				
						
	}
	else if( inList[inScanIndex].mType == ITEMTYPE_MM2HALF )
	{
		/* 前回のパケットと区別するためにウェイト期間 */
		_delay_ms(6);

		/* 二回目は、解除信号（ソレノイドへの励磁を完了させる）*/
		if( inList[inScanIndex].mCnt <= 1 )
		{
			inList[inScanIndex].mDatas[2] = 0x00;
		}
		
		/* 合計8パケット(ダブルパケットx4)を送る。間隔は762us、1.5ms。 */
		for( int i = 0; i < 4; i++)
		{
			MM2_SendPackets(inList[inScanIndex].mDatas, 3, MODEMM2_HALF, 0, 0);
			_delay_us(762);
			MM2_SendPackets(inList[inScanIndex].mDatas, 3, MODEMM2_HALF, 0, 0);
			_delay_us(1500);			
		}
		
	}	
	else if( inList[inScanIndex].mType == ITEMTYPE_DCC )
	{
		
		DCC_SendPackets(inList[inScanIndex].mDatas, inList[inScanIndex].mLen);
		
		/* 回数制限のあるパケットは二回送る */
		if( inList[inScanIndex].mCnt > 0)
		{
			DCC_SendPackets(inList[inScanIndex].mDatas, inList[inScanIndex].mLen);
		}		
	}
	
	else if( inList[inScanIndex].mType == ITEMTYPE_UNKNOWN )
	{
		aRetIndex = 255;
	}	
	
	else
	{
		aRetIndex = 255;
	}	
		
	/* カウント値が入っている指令は時間が来たら削除する */
	if( (inList[inScanIndex].mCnt > 0) & (aRetIndex != 255))
	{
		inList[inScanIndex].mCnt--;
		if( inList[inScanIndex].mCnt <= 0)
		{
			DSList_DeleteItem(inList, inScanIndex);
		}	
	}	
	
	return aRetIndex;
}



 
UC DSCoreLib::Main_decodeLocDirection(DS_ITEM *inList, US address, UC direction, UC *outType, UC *outPackets, UC *outLen)
{
	UC aAddr1, aAddr2;
	UC aLen;
	UC aType;
	UC aDirection;
	UC aIndex;
	
	/* decode loc address */
	Main_decodeAddress( address, &aType, &aAddr1, &aAddr2, &aLen);
	
	/* パケットのアドレス部分は確定するのでセットする */
	outPackets[0] = aAddr1;
	
	if( aLen >= 2)
	{
		outPackets[1] = aAddr2;
	}
	
	*outType = aType;	
	
	/* 前回のファンクションデータが必要なため同じアドレスを探索 */
	aIndex = DSList_FindItem(inList, aType, aLen, 0, outPackets);
	
	if( aIndex != 255)
	{
		aDirection = inList[aIndex].mDirection;
	}
	else
	{
		return 255;
	}
	
	if( aDirection != direction)
	{
		/* 異なる場合はリスト側を書き換える */
		inList[aIndex].mDirection = direction;
		return 1;		
	}
	else
	{
		/* 同じ場合は書き換える必要なし */
		return 0;
	}
	
}
 
void DSCoreLib::Main_decodeLocFunction(DS_ITEM *inList, US address, UC function, UC power, UC *outType, UC *outPackets, UC *outLen, UL *ioFuncBuf)
{
	UC aAddr1, aAddr2;
	UC aLen;
	UC aType;
	UL aFunctionBuf;
	UC aIndex;
	UC aOffset;

	/* decode loc address */
	Main_decodeAddress( address, &aType, &aAddr1, &aAddr2, &aLen);
	
	/* パケットのアドレス部分は確定するのでセットする */
	*outType = aType;	
	outPackets[0] = aAddr1;
	
	if( aLen >= 2)
	{
		outPackets[1] = aAddr2;
	}
				
	/* 前回のファンクションデータが必要なため同じアドレスを探索 */
	aIndex = DSList_FindItem(inList, aType, aLen, 0, outPackets);
	
	if( aIndex != 255)
	{
		aFunctionBuf = inList[aIndex].mFunctionBuf;
	}
	else
	{
		aFunctionBuf = 0;
	}				
				
	
	/* decode function  */
	
	switch( aType)
	{
		case ITEMTYPE_MM2BASE:
			outPackets[1] = 0b00;
			outPackets[2] = 0x00;/* ダミー */
			outPackets[3] = 0x00;
			
			/* 4バイト目にファンクション情報を記載しておく */
			if( power == 0)
			{
				*ioFuncBuf = aFunctionBuf & (~(1UL << (function - 1)));
			}
			else
			{
				*ioFuncBuf = aFunctionBuf | (1UL << (function - 1));
			}
			*outLen = 3;	

		break;
		
		case ITEMTYPE_DCC:
		
			aOffset = ( aLen >= 2) ? 1 : 0;

			/*　ファンクションの書き換えとDCCのパケット生成　*/
			outPackets[1 + aOffset] = DCC_decodeFunction(function, power, &aFunctionBuf);
				
			/* F13-F20, F21-28の拡張ファンクション用（ファンクション用にサイズが増える） */
			switch(outPackets[1 + aOffset])
			{
				case 0b11011110:
				outPackets[2 + aOffset] = ((aFunctionBuf >> 13) & 0xFF);
				*outLen = 3 + aOffset;
				break;
					
				case 0b11011111:
				outPackets[2 + aOffset] = ((aFunctionBuf >> 21) & 0xFF);
				*outLen = 3 + aOffset;
				break;
					
				default:
				*outLen = 2 + aOffset;
				break;
			}
			
			*ioFuncBuf = aFunctionBuf;
		
		
		break;
	}	
	
}

void DSCoreLib::Main_decodeLocSpeed(DS_ITEM *inList, US address, short speed, UC *outType, UC *outPackets, UC *outLen, UC inSpeedStep)
{
	UC aAddr1, aAddr2;
	UC aLen;
	UC aType;
	UC aDirection;
	UC aIndex;
	UC aSpeed;
	UC aOffset;

	/* decode loc address */
	Main_decodeAddress( address, &aType, &aAddr1, &aAddr2, &aLen);
	
	/* パケットのアドレス部分は確定するのでセットする */
	*outType = aType;
	outPackets[0] = aAddr1;
	
	if( aLen >= 2)
	{
		outPackets[1] = aAddr2;
	}	
	
	/* 前回のファンクションデータが必要なため同じアドレスを探索 */
	aIndex = DSList_FindItem(inList, aType, aLen, 0, outPackets);
	
	if( aIndex != 255)
	{
		aDirection = inList[aIndex].mDirection;
	}
	else
	{
		aDirection = 0;
	}
		
	/* Speedstepのセット */
	if( inSpeedStep != 0xFF)
	{
		inList[aIndex].mSpeedStep = inSpeedStep;
	}	
	
	/* decode speed  */
	
	switch( aType)
	{
	case ITEMTYPE_MM2BASE:
		outPackets[1] = 0b00;
		
		if(inList[aIndex].mSpeedStep == SPEEDSTEP_MM14)
		{
			/* 14 step speed */			
			aSpeed = DS_CalcSpeedStep(speed, 14);
			outPackets[2] = MM2_encodeSpeed14Step(aSpeed, aDirection);
			*outLen = 3;
		}
		else if(inList[aIndex].mSpeedStep == SPEEDSTEP_MM28)
		{
			/* 28 step speed */		
			UC aRawSpeed = DS_CalcSpeedStep(speed, 28);
			UC aSecondSpeed = aRawSpeed >> 1;
			
			if( aRawSpeed > (aSecondSpeed << 1))
			{
				aSpeed = aSecondSpeed + 1;
			}
			else
			{
				if( aRawSpeed > 0)
				{
					aSpeed = aSecondSpeed - 1;
				}
				else
				{
					aSpeed = 0;
				}
			}
			
			outPackets[2] = MM2_encodeSpeed14Step(aSpeed, aDirection);
			outPackets[3] = MM2_encodeSpeed14Step(aSecondSpeed, aDirection);			
			*outLen = 4;
		}
		else
		{
			outPackets[2] = 0;
			*outLen = 3;			
		}			
					
				
		break;
		
	case ITEMTYPE_DCC:
		
		aOffset = (aLen >= 2) ? 1: 0;
					
		if(inList[aIndex].mSpeedStep == SPEEDSTEP_DCC28)
		{
			/* 28step速度 */
			outPackets[1 + aOffset] = DCC_decodeSpeedStep28(speed, aDirection);
			*outLen = 2 + aOffset;
		}
		else if(inList[aIndex].mSpeedStep == SPEEDSTEP_DCC127)
		{
			/* 拡張127step速度 */
			outPackets[1 + aOffset] = 0b00111111;
			outPackets[2 + aOffset] = DCC_decodeSpeedStep127(speed, aDirection);
			*outLen = 3 + aOffset;
		}
		else if(inList[aIndex].mSpeedStep == SPEEDSTEP_DCC14)
		{
			/* 14step速度 */
			outPackets[1 + aOffset] = DCC_decodeSpeedStep14(speed, aDirection, (UC)(inList[aIndex].mFunctionBuf & 0b1));
			*outLen = 2 + aOffset;
		}
		else
		{
			outPackets[1 + aOffset] = 0;
			*outLen = 2 + aOffset;
		}
		
		break;
	}
	

}

void DSCoreLib::Main_decodeAccessory(US address, UC power, UC *outType, UC *outPackets, UC *outLen)
{
	UC aAddr1, aAddr2;
	UC aLen;
	UC aType;
	UC aPower;
	UC aProtocolType = DS_GetLocIDProtocol((UC)(address >> 8));
	
	/* decode position  */
	
	switch( aProtocolType)
	{
	case PROTOCOL_DCC_ACC:
		/* decode acc address (basic 9bit accessory mode) */
		Main_decodeAddress( address, &aType, &aAddr1, &aAddr2, &aLen);
		
		/* Power: 0 Straight(Green), 1 Diverging(Red) */
			
		*outType = ITEMTYPE_DCC;
		outPackets[0] = aAddr1;
		outPackets[1] = aAddr2 | 0b1000 | (power & 0b1) | ((((address - ADDR_ACC_DCC) % 4) & 0b11)  << 1);//1AAACDDD
		*outLen = 2;
	
		break;
		
	case PROTOCOL_MM2_ACC:
		/* decode acc address */
		Main_decodeAddress( address, &aType, &aAddr1, &aAddr2, &aLen);
		
		if( power > 0)
		{
			/* RED(1) */
			aPower = 0b11000000; //8,7bit目はActive信号
		}
		else
		{
			/* GREEN(0) */
			aPower = 0b11000011; //8,7bit目はActive信号
		}
		
		*outType = ITEMTYPE_MM2HALF;
		outPackets[0] = aAddr1;
		outPackets[1] = 0b00;
		outPackets[2] = MM2_encodeAccSecondAddress((address - 0x3000) % 4) | aPower;
		*outLen = 3;
		break;
	}
	

}


void DSCoreLib::Main_decodeAddress( US address, UC *outpType, UC *outpAddr1, UC *outpAddr2, UC *outpLen)
{
	US aTempAddr;
	UC aType = DS_GetLocIDProtocol((UC)(address >> 8));
	
	switch(aType)
	{
		case PROTOCOL_MM2:
			aTempAddr = address;
			
			if( aTempAddr <= 256)
			{
				*outpType = ITEMTYPE_MM2BASE;
				*outpAddr1 = MM2_encodeAddress(aTempAddr);
				*outpAddr2 = 0;
				*outpLen = 1;
			}
			else
			{
				*outpLen = 0;
			}
			break;

		case PROTOCOL_MM2_ACC:
			
			aTempAddr = ((address - 0x3000) >> 2) + 1;
			
			if( aTempAddr <= 80)
			{
				*outpType = ITEMTYPE_MM2HALF;
				*outpAddr1 = MM2_encodeAddress(aTempAddr);
				*outpAddr2 = 0;
				*outpLen = 1;
			}
			else
			{
				*outpLen = 0;
			}
			break;

		case PROTOCOL_DCC:
			*outpType = ITEMTYPE_DCC;
			
			/* offset DCC address to convert CAN address to DCC address */
			aTempAddr = address - 0xC000;

			if( aTempAddr <= 127)
			{
				/* 0-127 address */
				*outpAddr1 = (UC)(aTempAddr & 0b01111111);
				*outpAddr2 = 0;
				*outpLen = 1;
			}
			else if( aTempAddr < 10239)
			{
				/* DCC 14bit address range is 0b11000000-0b00000000(49152, addr.0) to 0b11100111-0b11111111(59391) */
				/* 128-10239 address, but 0-127 long address doesn't use */
				aTempAddr = aTempAddr + 49152;
				*outpAddr1 = (UC)((aTempAddr >> 8) & 0xFF);
				*outpAddr2 = (UC)(aTempAddr & 0xFF);
				*outpLen = 2;
			}
			else
			{
				/* out of 9bit/14bit locomotive address range */
				*outpLen = 0;
				
			}
			
			break;

		case PROTOCOL_DCC_ACC:
			*outpType = ITEMTYPE_DCC;
			
			/* offset DCC address to convert CAN address to DCC address */
			aTempAddr = ((address - ADDR_ACC_DCC) >> 2) + 1;
			
			if( aTempAddr < 512)
			{
				*outpAddr1 = (UC)(0b10000000 | (aTempAddr & 0b00111111));//10AAAAAA 
				*outpAddr2 = (UC)(0b10000000 | ((((~aTempAddr) & 0b111000000) >> 2) & 0xFF));//1AAACDDD
				*outpLen = 2;
			}
			else
			{
				/* out of 9bit accessory address range */
				//Extended Accessory Decoder Control Packet Format
				*outpLen = 0;
				
			}
			
			break;
		default:
			*outpLen = 0;
			break;
	}

}



void DSCoreLib::Main_registerDCCPoweronSequence()
{
	//アイテム登録用
	DS_ITEM aItem;	
	
	/* 全DCCデコーダリセット処理 */
	DCC_SendInitialPacket();
				
	/* DCC仕様上、20ms以上待つ。設定モードに移行させないため。 */
	_delay_ms(21);

	/* DCC Idle */
	aItem.mType = ITEMTYPE_DCC;
	aItem.mCmd = CMD_PWR_ON;
	aItem.mLen = 2;
	aItem.mCnt = 0;
	aItem.mDatas[0] = 0b11111111;
	aItem.mDatas[1] = 0b00000000;
	
	DSList_AddItem(gList, aItem);
}


void DSCoreLib::Init()
{

	/* ゲート信号制御 */
	DDRD =  (1<<DDD6)|(1<<DDD5);
	PORTD = 0b00011100;	/* INT0, INT1, PCINT20のプルアップをON */
	
	/* 初期化 */
	DCC_Init();
	
	/* リストスキャンを初期化 */
	gScanIndex = 0;
	DSList_Clear(gList);	
	
}


void DSCoreLib::Scan()
{
	/* リストのスキャン */
	if( MAIN_scanList(gList, gScanIndex) != 255)
	{
		/* 次のパケットへ */
		gScanIndex++;
	}
	else
	{
		/* リセット */
		gScanIndex = 0;
	}
}




boolean DSCoreLib::IsPower()
{
	return gPoweron;
}

boolean DSCoreLib::SetPower(boolean power)
{

	if (power) {
		/* 線路に電源投入 */
		DS_Power(1);
		
		/* デコーダ側の電源が安定するまで待つ */
		_delay_ms(10);
		
		/* リストをクリアする */
		//DSList_Clear(gList);
		
		 Main_registerDCCPoweronSequence();		
		
		//DEBUG
		//Serial.println("[CMD] Power On");
		gPoweron = true;
	}
	else
	{
		DS_Power(0);
		//DEBUG
		//Serial.println("[CMD] Power Off");
		
		gPoweron = false;
	}

	return true;
}


boolean DSCoreLib::SetLocoSpeed(word address, int inSpeed)
{
	return SetLocoSpeedEx( address, inSpeed, 0xFF);
}


boolean DSCoreLib::SetLocoSpeedEx(word address, int inSpeed, int inProtcol)
{
	UC aPackets[5] = {0,0,0,0,0};
	UC aLen;
	UC aType;
	
	/* アドレスデコード処理 */
	Main_decodeLocSpeed(gList, address, inSpeed, &aType, aPackets, &aLen, inProtcol);
	
	/* 管理リストをアップデート */
	DSList_UpdateItem(gList, aType, CMD_SPEED, aLen, 0, aPackets);
	
	//DEBUG
	/*
	Serial.print("[CMD] ADR=");
	Serial.print(aAddress);
	Serial.print(",SPD=");
	Serial.print(aSpeed);
	Serial.print(",PRT=");
	Serial.println(aProtocol);
	*/
} 

boolean DSCoreLib::SetLocoFunction(word address, unsigned char inFunction, unsigned char inPower)
{
	UC aPackets[5] = {0,0,0,0,0};
	UC aLen;
	UC aType;
	UL aFuncBuf = 0;
	
	
	/* ファンクション指令デコード */
	Main_decodeLocFunction(gList, address, inFunction, inPower, &aType, aPackets, &aLen, &aFuncBuf);

	/* ファンクション命令は上書き(MM2用, DCCは通常のアップデート) */
	DSList_UpdateItem_OWFunction(gList, aType, CMD_FUNCTION, aLen, aFuncBuf, inFunction, aPackets);

	return true;
}


boolean DSCoreLib::SetLocoDirection(word address, unsigned char inDirection)
{
	UC aPackets[5] = {0,0,0,0,0};
	UC aLen;
	UC aType;
	UC aDir = inDirection - 1;
	
	/* アドレスデコード処理,リストと指令が異なる場合は停止させる */
	switch( Main_decodeLocDirection(gList, address, aDir, &aType, aPackets, &aLen) )
	{
		case 1:
			/* 速度をゼロにして止める */
			Main_decodeLocSpeed(gList, address, 0, &aType, aPackets, &aLen, 0xFF);
			DSList_UpdateItem(gList, aType, CMD_SPEED, aLen, 0, aPackets);
			break;
		
		case 255:
			/* 新規登録処理 */
			Main_decodeLocSpeed(gList, address, 0, &aType, aPackets, &aLen, 0xFF);
			/* 登録リストに入っていないものは、リストに新規登録（進行方向を格納するため） */
			DSList_NewItem(gList, aType, CMD_SPEED, aLen, 0, aDir, aPackets);
			break;
		
		default:
			break;
	}
	
	return true;
}


boolean DSCoreLib::SetTurnout(word address, byte inSwitch)
{
	UC aPackets[5] = {0,0,0,0,0};
	UC aLen;
	UC aType;
	UC aPower;

	//inPackets[3];は未使用。（Positionの予約だったが、使わないことになってしまった。)
	aPower = convertAcc_MMDCC(address, inSwitch);
	
	/* アクセサリ指令デコード */
	Main_decodeAccessory(address, aPower, &aType, aPackets, &aLen);

	/* 管理リストをアップデート */
	DSList_UpdateItem(gList, aType, CMD_ACCESSORY, aLen, 2, aPackets);		

	return true;
}

byte DSCoreLib::convertAcc_MMDCC(word address, byte inSwitch)
{
	
	switch( GetLocIDProtocol(address >> 8))
	{
	case ADDR_ACC_MM2:
		/* 0:Straight, 1: diverging */
		return (inSwitch == 0) ? 1 : 0;
		break;
		
	case ADDR_ACC_DCC:
		/* 1:Straight, 0: diverging */
		return inSwitch;
		break;
		
	default:
		return inSwitch;
		break;
	}
	
}

word DSCoreLib::GetLocIDProtocol(byte address)
{
	if( address < 0x04)
	{
			return ADDR_MM2;
	}
	else if( (address >= 0x30) && (address <= 0x33))
	{
			return ADDR_ACC_MM2;
	}
	else if( (address >= 0x38) && (address <= 0x3F))
	{
			return ADDR_ACC_DCC;
	}
	else if( (address >= 0x40) && (address <= 0x70))
	{
			return ADDR_MFX;
	}
	else if( (address >= 0xC0) && (address <= 0xFF))
	{
			return ADDR_DCC;
	}
	else
	{
		return 0;
	}
}

boolean DSCoreLib::WriteConfig(word address, word number, byte value)
{
	UC aPackets[5] = {0,0,0,0,0};
	UC aLen;
	UC aType;
	word aOffsetCVNo;
	UC aPowerOnEnd;
	US aAddress;
	UC aCVValue;
	
	if( address >= ADDR_DCC)
	{
		aOffsetCVNo = number;
	}
	else
	{
		aOffsetCVNo = number | 0x8000;
	}

	/* CV書き込み */
	aAddress = aOffsetCVNo;
	aCVValue = value;
	
	aPowerOnEnd = DS_IsPowerOn();

	/* 電源遮断（電源オン時はいったん電源遮断） */
	DS_Power(0);
	_delay_ms(800);
		
	/* 線路に電源投入 */
	DS_Power(1);
		
	/* デコーダ側の電源が安定するまで待つ(ただし100ms未満であること。) */
	_delay_ms(50);
		
	/* 最上位ビットが立っているときはMM2, 0の場合はDCCとする */
	/* 0b10xx MM2
	   0b00xx DCC (標準)
	   0b01xx DCC (DS-DCCデコーダ専用, Bootloader起動完了待ち)
	  */
	
	
	if( (aAddress & 0x8000) > 0)
	{
		/* 少々待つ */
		_delay_ms(3);
		/*MM2*/
		aAddress = aAddress & 0x7FFF;
			
		for(int j = 0; j < 8; j++)
		{
			int aSendCount = 0;
				
			switch(j)
			{
				case 0:
				/* 機関車アドレスを120回送る(後半は11) */
				aPackets[0] = 0b00000000;
				aPackets[1] = 0b00;
				aPackets[2] = 0b00000011;
				aSendCount = 120;
				break;
					
				case 1:
				/* 機関車アドレスを30回送る(後半は00) */
				aPackets[2] = 0b00000000;
				aSendCount = 30;
				break;
					
				case 2:
				/* CV Noを送る(後半は00) */
				aPackets[0] = MM2_encodeAddress(aAddress);
				aPackets[2] = 0b00000000;
				break;
					
				case 3:
				/* CV Noを送る(後半は11) */
				aPackets[2] = 0b00000011;
				break;
				
				case 4:
				/* CV Noを送る(後半は00) */
				aPackets[2] = 0b00000000;
				break;					
					
				case 5:
				/* CV Valueを送る(後半は00) */
				aPackets[0] = MM2_encodeAddress(aCVValue);
				aPackets[2] = 0b00000000;
				break;
					
				case 6:
				/* CV Valueを送る(後半は11) */
				aPackets[2] = 0b00000011;
				break;
				
				case 7:
				/* CV Valueを送る(後半は00) */
				aPackets[2] = 0b00000000;
				aSendCount = 60;
				break;					
			}
				
			/* パケットを送る */
			for( int i = 0; i < aSendCount; i++)
			{
				MM2_SendDoublePackets(aPackets, 3, MODEMM2_BASE, 0, 0);
				MM2_WaitLong();
			}
		}

	}
	else
	{
		
		/* DS-DCC Decoder専用でブートローダの立ち上がり(1sec)を待つ */
		if(	(aAddress & 0x4000) > 0)
		{
			_delay_ms(3000);
			aAddress = aAddress & 0x3FFF;
		}
		
		/*DCC*/
		aAddress = aAddress - 1;/* 1-1023を、0-1022に変換 */
			
		/* リセットパルスを3回以上送る */
		for( int i = 0; i < 10; i++)
		{
			DCC_SendInitialPacket();
			_delay_us(150);
		}
			
		/* CVパケット生成(書き込み指令) */
		aPackets[0] = 0b01110000 | 0b00001100 | ((aAddress >> 8) & 0b11);
		aPackets[1] = aAddress & 0b11111111;
		aPackets[2] = aCVValue;
			
		/* CV指令を5回以上送る */
		for( int i = 0; i < 10; i++)
		{
			DCC_SendPacketsEx(aPackets, 3, 24);
			_delay_us(150);
		}
			
		/* リセットパルスを6回以上送る */
		for( int i = 0; i < 12; i++)
		{
			DCC_SendInitialPacket();
			_delay_us(150);
		}

	}

	/* いったん電源を切る */
	DS_Power(0);

		
	/* 電源遮断(最初から電源が入っているときは、そのままオン) */
	if( aPowerOnEnd != 0)
	{
		_delay_ms(1000);
		DS_Power(1);
		_delay_ms(1);
	}

	return true;
}




	
	